import os
from pathlib import Path
from typing import Dict, List

import pandas as pd
from sqlitedict import SqliteDict


class Utility:
    """Static class with utility methods.

    """
    
    @staticmethod
    def json_value_or_default(
        json_data: Dict[str, object],
        *keys: str,
        default=0
    ) -> object:
        """
        Check the json_data for a value associated with the provided key.  If no
        such key exists, return the default value instead.

        TODO: Use dictionary unpacking to manage multiple indexing.
        TODO: Update all the dictionary indexing with this method.

        Args:
            json_data (Dict[str, object]): Data to parse value from.
            default (int, optional): Default value to return if any failure occurs.
            Defaults to 0.

        Returns:
            object: The parsed value or default.
        """
        try:
            value = json_data
            for key in keys:
                value = value[key]
            return value
        except KeyError:
            # TODO: Log
            return default

    @staticmethod
    def print_table(
        table: List[List[str]],
        align: str="",
        hasHeader: bool=False,
        pad: int=2,
        isGrid: bool=False
    ):
        """
        Takes a 2 dimensional list of strings and prints the contents out in the form
        of a table.
        
        Values must be of type string or method will throw.

        Args:
            table (List[List[str]]): A list of lists of strings holding the data
            to be tabulated.
            align (str, optional): Specify cell alignment. Defaults to "".
            hasHeader (bool, optional): Specify if data contains a header.
            Defaults to False.
            pad (int, optional): Specify the amount of cell padding to use.
            Defaults to 2.
            isGrid (bool, optional): Specify if the grid should be printed.
            Defaults to False.
        """
        table = [row[:] for row in table] # copy table
        numRows, numCols = len(table), len(table[0]) # table size
        align = align.ljust(numCols,"L") # align left by default
        align = ["RC".find(c)+1 for c in align] # convert to index (?RC=012)
        widths = [max(len(row[col]) for row in table) for col in range(numCols)] # column widths

        # --- apply column widths with alignments ---
        if hasHeader: # header is centered
            for x in range(numCols):
                table[0][x] = table[0][x].center(widths[x])
        for y in range(hasHeader, numRows): # apply column alignments
            for x in range(numCols):
                c = table[y][x]
                table[y][x] = [c.ljust, c.rjust, c.center][align[x]](widths[x])

        # --- data for printing
        P = " "*pad
        LSEP,SEP,RSEP = "│"+P, P+"│"+P, P+"│"
        lines = ["─"*(widths[col]+pad*2) for col in range(numCols)]

        drawLine = [isGrid]*numRows
        drawLine[0]|=hasHeader
        drawLine[-1] = False
        if hasHeader or isGrid:
            gridLine = "├"+"┼".join(lines)+"┤" # if any(drawLine)

        # --- print rows ---
        print("┌"+"┬".join(lines)+"┐")
        for y in range(numRows):
            print(LSEP+SEP.join(table[y])+RSEP)
            if drawLine[y]:
                print(gridLine)
        print("└"+"┴".join(lines)+"┘")

    @staticmethod
    def get_db_path(path: Path) -> str:
        """Get database file full path.

        Args:
            path (Path): Path where database file should reside.

        Returns:
            str: Full path to database file.
        """
        return os.path.join(path, "NHLPredictor.sqlite")
    
    @staticmethod
    def split_save_try_pair(value: str) -> tuple[int, ...]:
        """Method to split columns using compound values separated by a known
        delimeter.
        
        TODO: We should generalize this.

        Args:
            value (str): Value to be split

        Returns:
            tuple[int, ...]: Tuple containing the separated values.
        """
        parts = str(value).split('/')
        parts = [int(part) for part in parts]

        return tuple(parts)

    @staticmethod
    def get_sqlitedict_tables(
        *names: str,
        path: Path,
        update_db: bool = False,
        read_only: bool = False
    ) -> Dict[str, SqliteDict]:
        """Get a dictionary of table names to SqliteDict database objects.

        Args:
            names (str): Database table names.
            path (Path): Path to location to find or save database file.
            update_db (bool, optional): Indicates if tables should be cleared first. Defaults to False.
            read_only (bool, optional): Indicates if tables should be editable. Defaults to False.

        Returns:
            Dict[str, SqliteDict]: Dictionary of database tables.
        """
        DBs = {}
        if update_db:
            # Empty the database on open
            flag = "w"
        else:
            # Open for read/write without modification
            flag = "c"
        if read_only:
            # Process readonly flag last as we want it to take precedence.
            flag = "r"

        for name in names:
            DBs[name] = SqliteDict(
                Utility.get_db_path(path),
                tablename=name,
                autocommit=True,
                flag=flag
            )
        return DBs
    
    def get_pandas_tables(
        *names: str,
        path: Path
    ) -> Dict[str, pd.DataFrame]:
        """_summary_

        Args:
            names (str): Database table names.
            path (Path): Path to location to find or save database file.

        Returns:
            Dict[str, pd.DataFrame]: Dictionary of table names to pandas
            DataFrames holding each table's data.
        """
        DBs = Utility.get_sqlitedict_tables(
            *names,
            path=path,
            read_only=True
        )
        for name in names:
            DBs[name] = pd.DataFrame(list(DBs[name].values()), index=list(DBs[name].keys()))
        return DBs


from enum import Enum

class Position(Enum):
    """
    Enumeration to map positions to numerical values. Numerical values match those
    used by the NHL API and should only be changed if the API changes.
    """
    C = 1       # Center
    L = 2       # Left wing
    R = 3       # Right wing
    D = 4       # Defenseman
    G = 5       # Goalie


from __future__ import annotations
from enum import Enum

import nhl_predictor.model.summarizers.average_player_summarizer as avsum
from nhl_predictor.model.summarizers.summarizer import Summarizer


class SummarizerTypes(str, Enum):
    """Enumeration of the supported summarizer types.
    """
    none = "none"
    average_player_summarizer = "average"

    @staticmethod
    def get_summarizer(summarizer_type: SummarizerTypes) -> Summarizer:
        """Build a summarizer instance from the specified summarizer type.

        Args:
            summarizer_type (SummarizerTypes): The type of summarizer to get.

        Raises:
            Exception: Throws if summarizer type cannot be found.

        Returns:
            Summarizer: A summarizer instance of the requested type.
        """
        match summarizer_type:
            case SummarizerTypes.average_player_summarizer:
                return avsum.AveragePlayerSummarizer()
            case _:
                # TODO: Shouldn't throw generic Exception
                raise Exception("Unsupported summarizer specified.")


from enum import Enum

class GameType(Enum):
    """
    Enum mapping game type labels to the numerical value used by the
    NHL API to indicate game type.

    Since the NHL doesn't document their API, value map was obtained
    from here:
    https://github.com/Zmalski/NHL-API-Reference/issues/23#issuecomment-2492925102
    """
    Preseason = 1
    RegularSeason = 2
    Playoff = 3

# The set of game types that are supported for prediction.
SupportedGameTypes = [
    GameType.RegularSeason,
    GameType.Playoff
]


import os
from pickle import load
from typing import List

import numpy as np
import pandas as pd

from nhl_predictor.model.home_or_away import HomeOrAway
from nhl_predictor.model.summarizer_manager import SummarizerTypes
from nhl_predictor.shared.constants.database import Database as DB
from nhl_predictor.shared.execution_context import ExecutionContext
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.shared.constants.json import JSON as Keys
from nhl_predictor.shared.utility import Utility as utl

logger = LoggingConfig.get_logger(__name__)
execution_context = ExecutionContext()
_summarizer = None
_model = None
_model_filename_part = "LinearRegression"

class PredictLinearRegression:

    @staticmethod
    def predict(
        games: List[int],
    ):
        """Predict the provided games.

        Args:
            games (List[int]): List of game IDs for the games to be predicted.
        """
        game_stats = pd.DataFrame()
        data = utl.get_pandas_tables(
            DB.players_table_name,
            DB.skater_stats_table_name,
            DB.goalie_stats_table_name,
            DB.games_table_name,
            path=execution_context.app_dir,
        )
        results_table = [["Away", "Home", "Predicted", "Raw"]]
        PredictLinearRegression._ensure_summarizer()
        PredictLinearRegression._ensure_model()

        if not games:
            # TODO: Check as precondition?
            logger.warning("No games on the schedule for chosen date(s).")
            return
        games = sorted(games, key=lambda item: item[Keys.id])
        game_stats = _summarizer.summarize_historical(games, data)
        if game_stats.empty:
            logger.warning("None of the specified games have released rosters yet.")
            # TODO: Probably need something here for usability.  Maybe print out
            # a message, an empty results table or even update logging to send
            # warnings to the console?
            print("None of the specified games have released rosters yet.")
            return
        data_pred = _model.predict(
            game_stats
            .sort_values(by=Keys.game_id)
            .drop(columns=Keys.game_id)
        )
        for i in range(len(data_pred)):
            home_team = utl.json_value_or_default(
                games[i],
                Keys.home_team,
                Keys.common_name,
                Keys.default
            )
            away_team = utl.json_value_or_default(
                games[i],
                Keys.away_team,
                Keys.common_name,
                Keys.default
            )
            prediction = HomeOrAway(np.rint(data_pred[i]).astype(int)).name
            results_table.append([away_team, home_team, prediction, str(data_pred[i])])

        utl.print_table(results_table, hasHeader=True)
    
    @staticmethod
    def _ensure_summarizer() -> None:
        """Ensures that the summarizer instance was created.
        """
        global _summarizer
        if _summarizer:
            pass
        else:
            _summarizer = SummarizerTypes.get_summarizer(
                execution_context.summarizer_type
            )
    
    @staticmethod
    def _ensure_model() -> None:
        """Make sure that the model has been loaded.
        
        TODO: Add some error checking here.
        """
        global _model
        if _model:
            pass
        if execution_context.model:
            model_filename = execution_context.model
        else:
            model_filename = f"{_summarizer.get_filename_prefix()}_{_model_filename_part}.pkl"
        with open(os.path.join(execution_context.app_dir, model_filename), "rb") as file:
            _model = load(file)


from abc import ABC, abstractmethod
from typing import Dict

import pandas as pd
from sqlitedict import SqliteDict

class Summarizer(ABC):
    """Represents an abstract summarizer that is used to process raw data from
    the database into a single Pandas DataFrame to be used by machine learning
    algorithm.
    
    Subclasses must implement the 'get_filename_prefix' and 'summarize' metods.
    """
    
    @abstractmethod
    def get_filename_prefix() -> str:
        """Provides a default prefix to use if the user does not specify an
        output file name. Prefix should be descriptive of the style used to
        reduce player stats.

        Returns:
            str: Prefix for model output filename.
        """
        pass

    @abstractmethod
    def summarize(
        self,
        data: Dict[str, SqliteDict]
    ) -> pd.DataFrame:
        """Summarizes stats into the final data set to be used by the learning
        module.

        Args:
            data (Dict[str, SqliteDict]): Dictionary containing the tables of
            raw data. Keys are table names as strings, each table is a SqliteDict.

        Returns:
            pd.DataFrame: Pandas DataFrame with the final data set for machine
            learning.
        """
        pass


from __future__ import annotations
from enum import Enum
from typing import List

class Seasons(str, Enum):
    """
    Seasons to use when fetching data from the API.  This is the data that we will
    be training and testing our model with.

    Since I only care about predicting Kraken performance so I only go back to their
    first year.  I might increase the scope in the future for better accuracy, but
    for now I value the smaller dataset for performance reasons.
    """
    _20222023 = "20222023"
    _20232024 = "20232024"
    _20242025 = "20242025"
    _20252026 = "20252026"

    @classmethod
    def items(cls) -> List[Seasons]:
        """Returns a list of all defined Seasons.

        Returns:
            List[Seasons]: The list of Seasons.
        """
        return [e for e in cls]


import os
from pathlib import Path

import typer
from nhlpy import NHLClient


class ExecutionContext:
    """A singleton execution context to track values needed throughout
    processing.
    """
    _app_name = "nhlpredictor"
    _app_dir_set = False
    
    def __new__(cls):
        """Overload __new__ to create only one instance.
        """
        if not getattr(cls, '_instance', None):
            cls._instance = super(type(ExecutionContext), cls).__new__(cls)
        return cls._instance

    @property
    def client(self) -> NHLClient:
        """Get the NHLClient for the application run.

        Returns:
            NHLClient: An NHLClient to use.
        """
        if getattr(self, '_client', None) is None:
            self._client = NHLClient()
        return self._client
    
    @property
    def summarizer_type(self) -> str:
        """Get the summarizer type for the application run.

        Returns:
            str: The summarizer type.
        """
        return self._summarizer

    @summarizer_type.setter
    def summarizer_type(self, value: str):
        """Set the summarizer type.

        Args:
            value (str): The value to set the summarizer type to.
        """
        self._summarizer = value

    @property
    def allow_update(self) -> bool:
        """Get if update is allowed.

        Returns:
            bool: Boolean indicating if update is allowed.
        """
        return self._allow_update

    @allow_update.setter
    def allow_update(self, value: bool):
        """Set if update is allowed.

        Args:
            value (bool): The value to set allow_update to.
        """
        self._allow_update = value
    
    @property
    def model(self) -> str:
        """Get the model file name.

        Returns:
            str: The model file name.
        """
        return self._model
    
    @model.setter
    def model(self, value: str):
        """Set the model file name.

        Args:
            value (str): The value to set the model file name.
        """
        self._model = value

    @property   
    def output_file(self) -> str:
        """Get the output file name.

        Returns:
            str: The output file name.
        """
        return self._output_file
    
    @output_file.setter
    def output_file(self, value: str):
        """Set the output file name.

        Args:
            value (str): The value to set the output file name to.
        """
        self._output_file = value

    @property
    def app_dir(self) -> Path:
        """Get the configured application directory.

        Returns:
            Path: The Path to the application directory.
        """
        if not getattr(self, "_app_dir", None):
            self.app_dir = typer.get_app_dir(ExecutionContext._app_name)
        return self._app_dir

    @app_dir.setter
    def app_dir(self, value: Path):
        """Set the application directory.

        Args:
            value (Path): Path to the application directory.

        Raises:
            Exception: Thrown if application directory is already set.
            RuntimeError: Thrown if path is invalid.
        """
        # TODO: shouldn't use general exceptions
        if ExecutionContext._app_dir_set:
            raise Exception("File path already set.")
        value = os.path.expanduser(value)
        if not os.path.isabs(value):
            value = os.path.abspath(value)
        if not os.path.isabs(value):
            raise RuntimeError("Error: could not resolve path.")
        if not os.path.exists(value):
            Path(value).mkdir()

        self._app_dir = value
        ExecutionContext._app_dir_set = True
    
    def _ensure_app_dir(self):
        """Method to ensure that the application directory has been initizlized.
        """
        self.app_dir


import logging

from pythonjsonlogger import jsonlogger


class LoggingConfig:
    """Class to manage logging configuration.
    """

    @staticmethod
    def get_logger(name: str):
        """Get a logger with the provided name.

        Args:
            name (str): Logger name.

        Returns:
            _type_: A properly configured logging object.
        """
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)

        # If logger with name was already created, don't muck with the handlers
        if not logger.handlers:
            handler = logging.FileHandler("buildData.log", 'w+')
            formatter = jsonlogger.JsonFormatter("%(asctime)s %(name)s %(levelname)s %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        return logger


import tensorflow as tf

"""
mnist = tf.keras.datasets.mnist

(x_train, y_train), (x_test, y_test) = mnist.load_data()
x_train, x_test = x_train / 255.0, x_test / 255.0

model = tf.keras.models.Sequential([
    tf.keras.layers.Flatten(input_shape=(28, 28)),
    tf.keras.layers.Dense(128, activation='relu'),
    tf.keras.layers.Dropout(0.2),
    tf.keras.layers.Dense(10)
])

predictions = model(x_train[:1]).numpy()
print(predictions)
"""

hello = tf.constant("hello, tensorflow")
sess = tf.compat.v1.Session()
print(sess.run(hello))


from pathlib import Path
from typing import List, Optional

import typer
from typing_extensions import Annotated

from nhl_predictor.model.algorithms import Algorithms
from nhl_predictor.model.seasons import Seasons
from nhl_predictor.model.summarizer_manager import SummarizerTypes
from nhl_predictor.shared.execution_context import ExecutionContext

# TODO: Review the CLI again.  Is there a pythonic way to do parameter sets?

app = typer.Typer()

# Option definition for specifying a summarizer.
_summarizer = Annotated[SummarizerTypes, typer.Option(
        help="Specify the algorithm to use to summarize roster strength.",
    )]

# Option definition for specifying a machine learning algorithm.
_algorithm = Annotated[Algorithms, typer.Option(
        help="Specify which ML algorithm to use.",
        case_sensitive=False,
    )]

# Option definition for specifying the application directory.
_app_dir = Annotated[Path, typer.Option(
    help=(
        "Specify the location to save related files. Default location is "
        "'~/.config/nhlpredictor'"
    )
)]

@app.command()
def build(
    season: Annotated[Optional[List[Seasons]], typer.Option(
        help=(
            "Specify the seasons to include. If '--all-seasons' is specified, "
            "this option will be ignored."
        )
    )] = None,
    all_seasons: Annotated[bool, typer.Option(
        help=(
            "Indicates that all seasons should be included in the data set. "
            "This option superscedes the '--season' option. See '--season' hints "
            "for list of seasons included."
        )
    )] = False,
    update: Annotated[bool, typer.Option(
        help=(
            "Existing tables will be cleared and repopulated."
        )
    )] = False,
    report: Annotated[bool, typer.Option(
        help=(
            "Reports on the current status of the database.  No alteration of "
            "of data will occur."
        )
    )] = False,
    app_dir: _app_dir = None
):
    """
    Build the data set.
    """
    context = ExecutionContext()
    context.allow_update = update
    if app_dir:
        context.app_dir = app_dir

    from nhl_predictor.builder.builder import Builder
    if report:
        Builder.report()
    else:
        Builder.build(season, all_seasons)

@app.command()
def train(
    algorithm: _algorithm = Algorithms.none,
    summarizer_type: _summarizer = SummarizerTypes.none,
    output: Annotated[str, typer.Option(
        help="Specify the file name for the serialized model."
    )] = None,
    update: Annotated[bool, typer.Option(
        help=(
            "Allow serialized model to be overwritten."
        )
    )] = False,
    app_dir: _app_dir = None
):
    """
    Train a model using the specified ML algorithm and data.
    """
    context = ExecutionContext()
    if app_dir:
        context.app_dir = app_dir
    context.output_file = output
    context.allow_update = update
    context.summarizer_type = summarizer_type
    
    from nhl_predictor.trainer.trainer import Trainer
    Trainer.train(algorithm)

@app.command()
def predict(
    algorithm: _algorithm = Algorithms.none,
    summarizer_type: _summarizer = SummarizerTypes.none,
    model: Annotated[str, typer.Option(
        help="Specify a pickle file containing the pre-trained model to use."
    )] = "",
    date: Annotated[str, typer.Option(
        help=(
            "Specify a date, all games for this date will be retrieved. This "
            "option superscedes the '--date-range' option."
        )
    )] = None,
    date_range: Annotated[str, typer.Option(
        help=(
            "Specify a range of dates, all games occuring during this range "
            "will be retrieved.  If '--date' is specified, this option will be "
            "ignored."
            "\n\nParsing of date range is performed by daterangeparser. See "
            "documentation for supported formats: "
            "https://daterangeparser.readthedocs.io/en/latest/\n"
            "\033[91m THIS IS NOT IMPLEMENTED YET.\033[91m"
            #TODO: Remove the not implemented disclaimer when appropriate
        )
    )] = None,
    list: Annotated[bool, typer.Option(
        help=(
            "Lists games based on the provided date or date range."
        )
    )] = False,
    game_id: Annotated[int, typer.Option(
        help=(
            "Specify a game to predict by its game ID."
        )
    )] = 0,
    app_dir: _app_dir = None
):
    """
    Predict the outcome of a game(s) given the specified model.
    """
    context = ExecutionContext()
    if app_dir:
        context.app_dir = app_dir
    context.model = model
    context.summarizer_type = summarizer_type
    
    from nhl_predictor.predictor.predictor import Predictor
    if list:
        Predictor.list_games(date, date_range)
    if game_id:
        Predictor.predict_single_game(algorithm, game_id)
    else:
        Predictor.predict_by_date(algorithm, date, date_range)

if __name__ == "__main__":
    """Main app entry point.
    """
    app()


class JSON:
    """Constants related to JSON
    """
    
    default = "default"
    last_update = "lastUpdate"

    # Game related keys.
    games = "games"
    id = "id"
    season = "season"
    game_type = "gameType"
    game_state = "gameState"
    player_by_game_stats = "playerByGameStats"
    home_team = "homeTeam"
    away_team = "awayTeam"
    forwards = "forwards"
    defense = "defense"
    goalies = "goalies"
    
    # Team related keys
    common_name = "commonName"

    # Stat related keys
    goals = "goals"
    assists = "assists"
    points = "points"
    plus_minus = "plusMinus"
    pim = "pim"
    hits = "hits"
    power_play_goals = "powerPlayGoals"
    sog = "sog"
    faceoff_winning_pctg = "faceoffWinningPctg"
    toi = "toi"
    blocked_shots = "blockedShots"
    shifts = "shifts"
    giveaways = "giveaways"
    takeaways = "takeaways"
    even_strength_saves_against = "evenStrengthSavesAgainst"
    even_strength_shots_against = "evenStrengthShotsAgainst"
    power_play_saves_against = "powerPlaySavesAgainst"
    power_play_shots_against = "powerPlayShotsAgainst"
    shorthanded_saves_against = "shorthandedSavesAgainst"
    shorthanded_shots_against = "shorthandedShotsAgainst"
    save_saves_against = "saveSavesAgainst"
    save_shots_against = "saveShotsAgainst"
    save_pctg = "savePctg"
    even_strength_goals_against = "evenStrengthGoalsAgainst"
    power_play_goals_against = "powerPlayGoalsAgainst"
    shorthanded_goals_against = "shorthandedGoalsAgainst"
    goals_against = "goalsAgainst"
    starter = "starter"
    decision = "decision"
    shots_against = "shotsAgainst"
    saves = "saves"

    # Player related keys
    player_id = "playerId"
    game_id = "gameId"
    is_Active = "isActive"
    first_name = "firstName"
    last_name = "lastName"
    current_team_id = "currentTeamId"
    height_in_cm = "heightInCentimeters"
    weight_in_kg = "weightInKilograms"

    # Custom keys. 
    # These are typically key names for the DB that don't exist in
    # the API JSON response formats.
    winner = "winner"
    home_team = "homeTeam"
    away_team= "awayTeam"
    score = "score"
    team_id = "teamId"
    team_role = "teamRole"
    skater_prefix = "skater_"
    goalie_prefix = "goalie_"
    home_suffix = "_home"
    away_suffix = "_away"


from typing import Dict

import pandas as pd
from sqlitedict import SqliteDict

from nhl_predictor.model.home_or_away import HomeOrAway
from nhl_predictor.model.summarizers.summarizer import Summarizer
from nhl_predictor.shared.constants.database import Database as DB
from nhl_predictor.shared.constants.json import JSON as Keys
from nhl_predictor.shared.execution_context import ExecutionContext
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.shared.utility import Utility as utl

logger = LoggingConfig.get_logger(__name__)
execution_context = ExecutionContext()

class AveragePlayerSummarizer(Summarizer):
    """Summarizer that simply averages or sums, as appropriate, each statistic
    across a game roster.
    """

    def get_filename_prefix() -> str:
        """Returns a prefix for naming the model save file with.

        Returns:
            str: File name prefix.
        """
        return "AverageSummarizer"
    
    def summarize(
        self,
        data: Dict[str, SqliteDict]
    ) -> pd.DataFrame:
        """Creates a single DataFrame containing the summarized data set.

        Args:
            data (Dict[str, SqliteDict]): Collection of raw data.

        Returns:
            pd.DataFrame: DataFrame with one row per game of summarized statistics.
        """
        self._cleanup_data(data)
        return self._reduce_data(data)
    
    def summarize_historical(
        self,
        games: Dict[str, object],
        data: Dict[str, SqliteDict]
    ) -> pd.DataFrame:
        """Summarize the historical player data for players participating in the
        provided games.

        Args:
            games (Dict[str, object]): Set of games to process.
            data (Dict[str, SqliteDict]): Collection of raw data.

        Returns:
            pd.DataFrame: Dataframe with one row per game of summarized statistics.
        """
        
        self._cleanup_data(data)
        # TODO: Can/should we do this in the cleanup method?
        data[DB.skater_stats_table_name] = data[DB.skater_stats_table_name].set_index(Keys.game_id)
        data[DB.goalie_stats_table_name] = data[DB.goalie_stats_table_name].set_index(Keys.game_id)
        
        game_stats = None
        
        for game in games:
            logger.info(f"Processing game. ID: '{utl.json_value_or_default(game, Keys.id)}'.")
            box_score = execution_context.client.game_center.boxscore(
                utl.json_value_or_default(game, Keys.id)
            )
            home_skater_ids = set(
                [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.home_team][Keys.forwards]]
                + [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.home_team][Keys.defense]]
            )
            away_skater_ids = set(
                [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.away_team][Keys.forwards]]
                + [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.away_team][Keys.defense]]
            )
            home_goalie_ids = set(
                [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.home_team][Keys.goalies]]
            )
            away_goalie_ids = set(
                [item[Keys.player_id] for item in box_score[Keys.player_by_game_stats][Keys.away_team][Keys.goalies]]
            )
            
            new_row = self._flatten_game_players(
                data,
                home_skater_ids,
                home_goalie_ids,
                away_skater_ids,
                away_goalie_ids
            )#.astype(object)
            new_row[Keys.game_id] = (utl.json_value_or_default(game, Keys.id))
            
            if game_stats is None:
                game_stats = pd.DataFrame([new_row])
            else:
                game_stats.loc[len(game_stats)] = new_row
        
        return game_stats.astype({Keys.game_id: 'int64'})

    def _cleanup_data(
        self,
        data: Dict[str, SqliteDict]
    ) -> None:
        """Cleans up the data to prepare for processing.

        Args:
            data (Dict[str, SqliteDict]): Collection of raw data.
        """
        skaters_db = data[DB.skater_stats_table_name]
        goalies_db = data[DB.goalie_stats_table_name]

        # TODO: Fix build-stage bug introducing duplicates then remove this.
        skaters_db = skaters_db.groupby([Keys.game_id, Keys.player_id]).first().reset_index()
        goalies_db = goalies_db.groupby([Keys.game_id, Keys.player_id]).first().reset_index()
        
        skaters_db = self._fix_skater_column_dtypes(skaters_db)
        self._split_compound_goalie_stats(goalies_db)
        goalies_db = self._fix_goalie_column_dtypes(goalies_db)

        data[DB.skater_stats_table_name] = skaters_db
        data[DB.goalie_stats_table_name] = goalies_db

    def _fix_skater_column_dtypes(
        self,
        skaters: pd.DataFrame
    ) -> pd.DataFrame:
        """Corrects the column dtypes for skater data.

        Args:
            skaters (pd.DataFrame): Raw skater data.

        Returns:
            pd.DataFrame: DataFrame with corrected column dtypes.
        """
        return skaters.astype({
            Keys.toi: "string",
        })
    
    def _fix_goalie_column_dtypes(
        self,
        goalies: pd.DataFrame
    ) -> pd.DataFrame:
        """Corrects the column dtypes for goalie data.

        Args:
            goalies (pd.DataFrame): Raw goalie data.

        Returns:
            pd.DataFrame: DataFrame with corrected column dtypes.
        """
        return goalies.astype({
            Keys.even_strength_shots_against: int,
            Keys.power_play_shots_against: int,
            Keys.shorthanded_shots_against: int,
            Keys.save_shots_against: int,
            Keys.toi: "string",
            Keys.decision: "string",
            Keys.even_strength_saves_against: int,
            Keys.power_play_saves_against: int,
            Keys.shorthanded_saves_against: int,
            Keys.save_saves_against: int
        })
    
    # The shot columns provide us with shots againsts, saves against and goals
    # against. Since saves + goals = shots, this may be too much information.
    # There are several columns like this that may be introducing duplicates
    # when split.
    # TODO: Need to unpack if these relationships are bad for the model.
    def _split_compound_goalie_stats(
        self,
        goalies: pd.DataFrame
    ) -> None:
        """Splits columns with compound values into individual columns. Changes are made in place.

        Args:
            goalies (pd.DataFrame): Raw goalie data.
        """
        goalies[[Keys.even_strength_saves_against, Keys.even_strength_shots_against]] = \
            goalies[Keys.even_strength_shots_against].str.split('/', expand=True)
        goalies[[Keys.power_play_saves_against, Keys.power_play_shots_against]] = \
            goalies[Keys.power_play_shots_against].str.split('/', expand=True)
        goalies[[Keys.shorthanded_saves_against, Keys.shorthanded_shots_against]] = \
            goalies[Keys.shorthanded_shots_against].str.split('/', expand=True)
        goalies[[Keys.save_saves_against, Keys.save_shots_against]] = \
            goalies[Keys.save_shots_against].str.split('/', expand=True)

    def _reduce_data(
        self,
        data: Dict[str, SqliteDict]
    ) -> pd.DataFrame:
        """Reduce player stats down to roster stats.

        Args:
            data (Dict[str, SqliteDict]): Collection of raw data.

        Returns:
            pd.DataFrame: DataFrame containing the stats reduced to roster granularity.
        """
        game_stats = self._flatten_all_stats(data)
        game_stats = self._add_wins_column(data, game_stats)
        return game_stats
    
    def _flatten_all_stats(
        self,
        data: Dict[str, SqliteDict]
    ) -> pd.DataFrame:
        """Completely flatten player stats into roster summarized stats.

        Args:
            data (Dict[str, SqliteDict]): Collection of raw data.

        Returns:
            pd.DataFrame: DataFrame with fully flattened stats.
        """
        skaters_db = data[DB.skater_stats_table_name]
        skaters_reduced = self._group_and_flatten_skaters_by_game(skaters_db)
        skaters_reduced = self._flatten_home_and_away_by_game(skaters_reduced, Keys.skater_prefix)
        
        goalies_db = data[DB.goalie_stats_table_name]
        goalies_reduced = self._group_and_flatten_goalies_by_game(goalies_db)
        goalies_reduced = self._flatten_home_and_away_by_game(goalies_reduced, Keys.goalie_prefix)        

        return pd.merge(
            skaters_reduced,
            goalies_reduced,
            how='outer',
            on=Keys.game_id
        )
        
    def _flatten_game_players(
        self,
        data: Dict[str, pd.DataFrame],
        home_skater_ids: set[int],
        home_goalie_ids: set[int],
        away_skater_ids: set[int],
        away_goalie_ids: set[int]
    ) -> pd.Series:
        """Flatten player stats for the provided players.

        Args:
            data (Dict[str, pd.DataFrame]): Collection of raw data.
            home_skater_ids (set[int]): Home skater player IDs.
            home_goalie_ids (set[int]): Home goalie player IDs.
            away_skater_ids (set[int]): Away skater player IDs.
            away_goalie_ids (set[int]): Away goalie player IDs.

        Returns:
            pd.Series: Series containing the summarized stats for this game.
        """
        skater_db = data[DB.skater_stats_table_name]
        goalie_db = data[DB.goalie_stats_table_name]
        
        home_skater_df = self._group_and_flatten_skaters_by_player(skater_db[skater_db[Keys.player_id].isin(home_skater_ids)]).sum()
        away_skater_df = self._group_and_flatten_skaters_by_player(skater_db[skater_db[Keys.player_id].isin(away_skater_ids)]).sum()
        home_goalie_df = self._group_and_flatten_goalies_by_player(goalie_db[goalie_db[Keys.player_id].isin(home_goalie_ids)]).sum()
        away_goalie_df = self._group_and_flatten_goalies_by_player(goalie_db[goalie_db[Keys.player_id].isin(away_goalie_ids)]).sum()
        
        home_skater_df = home_skater_df.add_suffix(Keys.home_suffix).add_prefix(Keys.skater_prefix)
        away_skater_df = away_skater_df.add_suffix(Keys.away_suffix).add_prefix(Keys.skater_prefix)
        home_goalie_df = home_goalie_df.add_suffix(Keys.home_suffix).add_prefix(Keys.goalie_prefix)
        away_goalie_df = away_goalie_df.add_suffix(Keys.away_suffix).add_prefix(Keys.goalie_prefix)
        
        return pd.concat([
            home_skater_df,
            away_skater_df,
            home_goalie_df,
            away_goalie_df
        ])

    def _add_wins_column(
        self,
        data: Dict[str, SqliteDict],
        game_stats: pd.DataFrame
    ) -> pd.DataFrame:
        """Add winner column to the data set DataFrame

        Args:
            data (Dict[str, SqliteDict]): Collection of raw data.
            game_stats (pd.DataFrame): Data set DataFrame

        Returns:
            pd.DataFrame: DataFrame joining the data set DataFrame with the winner data.
        """
        games_db = data[DB.games_table_name]
        games_db.index.name = Keys.game_id # TODO: We should fix this on the build side
        wins = pd.DataFrame(games_db[Keys.winner])
        wins.index = wins.index.astype(int)
        return pd.merge(
            game_stats,
            wins,
            how='left',
            on=Keys.game_id
        )
    
    def _group_and_flatten_skaters_by_player(
        self,
        skaters_db: pd.DataFrame
    ) -> pd.DataFrame:
        """Groups stats entries by player and then flattens skater stats into a single stat line.

        Args:
            skaters_db (pd.DataFrame): Skater data.

        Returns:
            pd.DataFrame: DataFrame with skater stats summarized by game
        """
        skaters_grouped = skaters_db.groupby([Keys.player_id])
        return self._flatten_skater_stats(skaters_grouped, 'mean').set_index(Keys.player_id)
    
    def _group_and_flatten_skaters_by_game(
        self,
        skaters_db: pd.DataFrame
    ) -> pd.DataFrame:
        """Groups stats entries by game and then flattens skater stats into single stat line.

        Args:
            skaters_db (pd.DataFrame): Skater data.

        Returns:
            pd.DataFrame: DataFrame with skater stats summarized by game.
        """
        skaters_grouped = skaters_db.groupby([Keys.game_id, Keys.team_id, Keys.team_role])
        return self._flatten_skater_stats(skaters_grouped)
    
    def _flatten_skater_stats(
        self,
        skaters_db: pd.DataFrame,
        method: str = 'sum'
    ) -> pd.DataFrame:
        """Aggregate skater stats into a single DataFrame

        Args:
            skaters_db (pd.DataFrame): DataFrame to aggregate
            method (str, optional): Specifies the agg method. Defaults to 'sum'.

        Returns:
            pd.DataFrame: Aggregated skater stats.
        """
        return skaters_db.agg({
            Keys.goals: method,
            Keys.assists: method,
            Keys.points: method,
            Keys.plus_minus: method,
            # PIM TODO
            Keys.hits: method,
            Keys.power_play_goals: method,
            Keys.sog: method,
            # faceoffWinningPctg TODO
            # TOI TODO
            Keys.blocked_shots: method,
            Keys.shifts: method,
            Keys.giveaways: method,
            Keys.takeaways: method,
        }).reset_index()
        
    def _group_and_flatten_goalies_by_player(
        self,
        goalies_db: pd.DataFrame
    ) -> pd.DataFrame:
        """Groups stats entries by player and then flattens goalie stats into single stat line.

        Args:
            goalies_db (pd.DataFrame): Goalie data.

        Returns:
            pd.DataFrame: DataFrame with goalie stats summarized by game.
        """
        goalies_grouped = goalies_db.groupby([Keys.player_id])
        return self._flatten_goalie_stats(goalies_grouped, 'mean').set_index(Keys.player_id)
        
    def _group_and_flatten_goalies_by_game(
        self,
        goalies_db: pd.DataFrame
    ) -> pd.DataFrame:
        """Groups stats entries by game and then flattens goalie stats into single stat line.

        Args:
            goalies_db (pd.DataFrame): Goalie data.

        Returns:
            pd.DataFrame: DataFrame with goalie stats summarized by game.
        """
        goalies_grouped = goalies_db.groupby([Keys.game_id, Keys.team_id, Keys.team_role])
        return self._flatten_goalie_stats(goalies_grouped)
    
    def _flatten_goalie_stats(
        self,
        goalies_db: pd.DataFrame,
        method: str = 'sum'
    ) -> pd.DataFrame:
        """Aggregate goalie stats into a single DataFrame

        Args:
            goalies_db (pd.DataFrame): DataFrame to aggregate
            method (str, optional): Specifies the agg method. Defaults to 'sum'.

        Returns:
            pd.DataFrame: Aggregated goalie stats.
        """
        return goalies_db.agg({
            Keys.even_strength_shots_against: method,
            Keys.power_play_shots_against: method,
            Keys.shorthanded_shots_against: method,
            Keys.save_shots_against: method,
            #Keys.save_pctg: goalies_grouped.apply(lambda x: (x[Keys.saves]/x[Keys.shots_against])) TODO
            Keys.even_strength_goals_against: method,
            Keys.power_play_goals_against: method,
            Keys.shorthanded_goals_against: method,
            Keys.pim: method,
            Keys.goals_against: method,
            #TOI TODO
            #starter TODO
            #decision TODO
            Keys.shots_against: method,
            Keys.saves: method,
            Keys.even_strength_saves_against: method,
            Keys.power_play_saves_against: method,
            Keys.shorthanded_saves_against: method,
            Keys.save_saves_against: method,
        }).reset_index()
        
    def _flatten_home_and_away_by_player(
        self,
        home: pd.DataFrame,
        away: pd.DataFrame,
        prefix: str,
    ) -> pd.DataFrame:
        """Combines the HOME and AWAY stats into a single game stats line using the
        provided prefix.

        Args:
            home (pd.DataFrame): The home stats.
            away (pd.DataFrame): The away stats.
            prefix (str): Prefix to add to column names.

        Returns:
            pd.DataFrame: DataFrame with HOME and AWAY rows joined.
        """
        return pd.merge(
            home,
            away,
            how='outer',
            on=Keys.player_id,
            suffixes=(Keys.home_suffix, Keys.away_suffix)
        ).set_index(Keys.player_id).add_prefix(prefix)

    def _flatten_home_and_away_by_game(
        self,
        data: pd.DataFrame,
        prefix: str
    ) -> pd.DataFrame:
        """Splits provided data by HOME or AWAY designation and combines the stats
        into a single game line using provided prefix.

        Args:
            data (pd.DataFrame): DataFrame of data to be flattened.
            prefix (str): Stat type column name prefix.

        Returns:
            pd.DataFrame: DataFrame with HOME and AWAY rows joined.
        """
        home = (
            data[data[Keys.team_role] == HomeOrAway.HOME.value]
            .drop([Keys.team_role, Keys.team_id], axis=1)
        )
        away = (
            data[data[Keys.team_role] == HomeOrAway.AWAY.value]
            .drop([Keys.team_role, Keys.team_id], axis=1)
        )
        return pd.merge(
            home,
            away,
            how='outer',
            on=Keys.game_id,
            suffixes=(Keys.home_suffix, Keys.away_suffix)
        ).set_index(Keys.game_id).add_prefix(prefix)


import json
from datetime import datetime, timezone
from typing import Dict, List

import requests
from ansimarkup import ansiprint as print
from sqlitedict import SqliteDict

from nhl_predictor.model.game_state import GameState, GameStatesForDataset
from nhl_predictor.model.game_type import GameType, SupportedGameTypes
from nhl_predictor.model.home_or_away import HomeOrAway
from nhl_predictor.model.seasons import Seasons
from nhl_predictor.model.team_map import TeamMap
from nhl_predictor.shared.constants.database import Database as DB
from nhl_predictor.shared.constants.json import JSON as Keys
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.shared.utility import Utility as utl

logger = LoggingConfig.get_logger(__name__)
execution_context = shared.execution_context.ExecutionContext()

class Builder:
    """Static class with API to fetch desired data from the public NHL API and
    save it in local databases.
    """
    
    @staticmethod
    def build(
        seasons: List[Seasons],
        all_seasons: bool = False
    ) -> None:
        """Main entry point for the builder.

        Args:
            seasons (List[Seasons]): List of seasons to include in the data set.
            all_seasons (bool, optional): Specify if all seasons should be included. Defaults to False.
        """
        logger.info("Call to build starting.")
        execution_context._ensure_app_dir()
        data = utl.get_sqlitedict_tables(
            DB.players_table_name,
            DB.skater_stats_table_name,
            DB.goalie_stats_table_name,
            DB.games_table_name,
            DB.meta_table_name,
            path=execution_context.app_dir,
            update_db=execution_context.allow_update
        )
        
        if all_seasons:
            Builder._build_stats_by_season(data)
        elif seasons is not None:
            Builder._build_stats_by_season(data, seasons)
        else:
            logger.error("Invalid season specification, cannot build data set.")
        Builder.populate_players(data)
        logger.info("Call to build is complete.")
    
    @staticmethod
    def report() -> None:
        """Prints a report on the current state of the raw data.
        """
        logger.info("Start dataset report.")
        execution_context._ensure_app_dir()
        data = utl.get_sqlitedict_tables(
            DB.players_table_name,
            DB.skater_stats_table_name,
            DB.goalie_stats_table_name,
            DB.games_table_name,
            DB.meta_table_name,
            path=execution_context.app_dir,
            read_only=True
        )
        games_db = data[DB.games_table_name]
        players_db = data[DB.players_table_name]
        skaters_db = data[DB.skater_stats_table_name]
        goalies_db = data[DB.goalie_stats_table_name]
        meta_db = data[DB.meta_table_name]

        # Summarize the games table
        games_summary_table = []
        games_summary_table.append(["Num. Rows", str(len(games_db))])
        games_summary_table.append([
            "Last Updated",
            str(meta_db[DB.games_table_name][Keys.last_update])
        ])

        # Summarize the players table
        players_summary_table = []
        players_summary_table.append(["Num. Rows", str(len(players_db))])
        players_summary_table.append([
            "Last Updated",
            str(meta_db[DB.players_table_name][Keys.last_update])
        ])

        # Summarize the skaters table
        skaters_summary_table = []
        skaters_summary_table.append(["Num. Rows", str(len(skaters_db))])
        skaters_summary_table.append([
            "Last Updated",
            str(meta_db[DB.skater_stats_table_name][Keys.last_update])
        ])

        # Summarize the goalies table
        goalies_summary_table = []
        goalies_summary_table.append(["Num. Rows", str(len(goalies_db))])
        goalies_summary_table.append([
            "Last Updated",
            str(meta_db[DB.goalie_stats_table_name][Keys.last_update])
        ])
        
        # Print all the tables
        print("\n<b><green>GAMES:</green></b>")
        print("<blue>The total number of historical games processed.</blue>")
        utl.print_table(games_summary_table)
        print("\n<b><green>PLAYERS:</green></b>")
        print("<blue>The number of unique players encountered during processing.</blue>")
        utl.print_table(players_summary_table)
        print("\n<b><green>SKATERS:</green></b>")
        print("<blue>The number of skater stat records.  This should be appx. <num_games> * 36</blue>")
        utl.print_table(skaters_summary_table)
        print("\n<b><green>GOALIES:</green></b>")
        print("<blue>The number of goalie stat records.  This should be appx. <num_games> * 4</blue>")
        utl.print_table(goalies_summary_table)
        print("\n")
        print(f"<magenta>Note: Current time UTC is: {datetime.now(timezone.utc)}</magenta>")
        print("\n")

        logger.info("Finished dataset report.")

    @staticmethod
    def _build_stats_by_season(
        data: Dict[str, SqliteDict],
        seasons: List[str] = [x.value for  x in Seasons.items()],
    ) -> None:
        """Iterates over the specified seasons and adds those seasons' data to
        the local database.

        Args:
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data in.
            seasons (List[str], optional): List of seasons to process. Defaults to all items in Seasons enumeration.
        """
        logger.info("Start building seasons.")
        for season in seasons:
            logger.info(f"Start of processing for season '{season}'.")
            for team in TeamMap:
                logger.info(f"Start processing for team '{team}' in season '{season}'.")
                try:
                    games_raw = execution_context.client.schedule.team_season_schedule(team, season)[Keys.games]
                    logger.info(f"Found '{len(games_raw)}' games for team '{team}' in season '{season}'.")
                    Builder._process_raw_games(games_raw, data)
                except Exception as e:
                    print("<red>Exception occured. Check logs.</red>")
                    logger.exception(
                        f"Exception processing team_season_schedule query. "
                        f"Games: '{json.dumps(games_raw, indent=4)}',"
                        f"Exception: '{str(e)}'.",
                        stack_info=True)
        logger.info("Finished building seasons.")
    
    @staticmethod
    def _process_raw_games(
        games_raw: Dict[str, object],
        data: Dict[str, SqliteDict]
    ) -> None:
        """Iterates over the provided raw game data and adds it to the local
        database.

        Args:
            games_raw (Dict[str, object]): Dictionary with raw game JSON.
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data
            in.
        """
        logger.info("Start processing game.")
        games_db = data[DB.games_table_name]
        meta_db = data[DB.meta_table_name]

        try:
            for game in games_raw:
                logger.info(f"Processing game: '{game}'.")
                try:
                    if (GameType(utl.json_value_or_default(game, Keys.game_type, default=GameType.Preseason))
                        not in SupportedGameTypes):
                        logger.info(
                            f"Skipping game '{game[Keys.id]}' which is not a "
                            f"supported game type. Type: '{game[Keys.game_type]}'."
                        )
                        continue
                    if (GameState(utl.json_value_or_default(game, Keys.game_state, default=GameState.Future))
                        not in GameStatesForDataset):
                        logger.info(
                            f"Skipping game '{game[Keys.id]}' which is not a "
                            f"supported game state. State: '{game[Keys.game_state]}'."
                        )
                        continue
                    if game[Keys.home_team][Keys.score] > game[Keys.away_team][Keys.score]:
                        winner = HomeOrAway.HOME.value
                    else:
                        winner = HomeOrAway.AWAY.value
                    # game ID is the primary key for the games DB
                    games_db[game[Keys.id]] = {
                        Keys.season: game[Keys.season],
                        Keys.game_type: game[Keys.game_type],
                        Keys.game_state: game[Keys.game_state],
                        Keys.home_team: game[Keys.home_team][Keys.id],
                        Keys.away_team: game[Keys.away_team][Keys.id],
                        Keys.winner: winner
                    }
                except Exception as e:
                    print("\033[31mException occured. Check logs.\033[0m")
                    logger.exception(
                        f"Exception adding game data to database. Exception: "
                        f"'{str(e)}'.",
                        stack_info=True
                    )

                try:
                    box_score = execution_context.client.game_center.boxscore(game[Keys.id])
                    Builder._process_box_score(box_score, data)
                except Exception as e:
                    print("\033[31mException occured. Check logs.\033[0m")
                    logger.exception(
                        f"Exception processing box_score query. Exception: "
                        f"'{str(e)}', box_score: '{json.dumps(box_score, indent=4)}'.",
                        stack_info=True
                    )

            meta_db[DB.games_table_name] = {
                Keys.last_update: datetime.now(timezone.utc)
            }
        except Exception as e:
            print("\033[31mException occured. Check logs.\033[0m")
            logger.exception(
                f"Exception processing team_season_schedule query. Exception: "
                f"'{str(e)}', games: '{json.dumps(games_raw, indent=4)}', "
                f"box_score: '{json.dumps(box_score, indent=4)}'.", stack_info=True
            )
        logger.info("Finished processing game.")

    @staticmethod
    def _process_box_score(
        box_score: Dict[str, object],
        data: Dict[str, SqliteDict]
    ) -> None:
        """Iterates of the provided raw box score data and adds it to the local
        database.

        Args:
            box_score (Dict[str, object]): Dictionary with raw box score JSON.
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data in.
        """
        logger.info("Processing box_score. BoxScore: '{box_score}'.")
        
        if Keys.player_by_game_stats not in box_score:
            logger.warning("Roster not published yet")
            return None
        
        home_team = utl.json_value_or_default(box_score, Keys.player_by_game_stats, Keys.home_team)
        away_team = utl.json_value_or_default(box_score, Keys.player_by_game_stats, Keys.away_team)

        Builder._process_skaters(
            home_team[Keys.forwards] + home_team[Keys.defense],
            data,
            utl.json_value_or_default(box_score, Keys.id),
            utl.json_value_or_default(box_score, Keys.home_team, Keys.id),
            HomeOrAway.HOME
        )
        Builder._process_goalies(
            home_team[Keys.goalies],
            data,
            utl.json_value_or_default(box_score, Keys.id),
            utl.json_value_or_default(box_score, Keys.home_team, Keys.id),
            HomeOrAway.HOME
        )
        Builder._process_skaters(
            away_team[Keys.forwards] + away_team[Keys.defense],
            data,
            utl.json_value_or_default(box_score, Keys.id),
            utl.json_value_or_default(box_score, Keys.away_team, Keys.id),
            HomeOrAway.AWAY
        )
        Builder._process_goalies(
            away_team[Keys.goalies],
            data,
            utl.json_value_or_default(box_score, Keys.id),
            utl.json_value_or_default(box_score, Keys.away_team, Keys.id),
            HomeOrAway.AWAY
        )

        logger.info("Box score processed.")

    @staticmethod
    def _process_skaters(
        skaters: Dict[str, object],
        data: Dict[str, SqliteDict],
        game_id: str,
        team_id: str,
        team_role: HomeOrAway
    ) -> None:
        """Process the skaters for a given game

        Args:
            skaters (Dict[str, object]): Skater JSON data.
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data in.
            game_id (str): Game ID for the game represented in the data.
            team_id (str): Team ID for the team represented in the data.
            team_role (HomeOrAway): Team role represented in the data.
        """
        logger.info("Started adding skaters to database.")
        skater_stats_db = data[DB.skater_stats_table_name]
        meta_db = data[DB.meta_table_name]

        for skater in skaters:
            logger.info(f"Processing skater. Skater:'{skater}'.")
            skater_stats_db[len(skater_stats_db)+1] = {
                Keys.game_id: game_id,
                Keys.player_id: utl.json_value_or_default(skater, Keys.player_id),
                Keys.goals: utl.json_value_or_default(skater, Keys.goals),
                Keys.assists: utl.json_value_or_default(skater, Keys.assists),
                Keys.points: utl.json_value_or_default(skater, Keys.points),
                Keys.plus_minus: utl.json_value_or_default(skater, Keys.plus_minus),
                Keys.pim: utl.json_value_or_default(skater, Keys.pim),
                Keys.hits: utl.json_value_or_default(skater, Keys.hits),
                Keys.power_play_goals: utl.json_value_or_default(skater, Keys.power_play_goals),
                Keys.sog : utl.json_value_or_default(skater, Keys.sog),
                Keys.faceoff_winning_pctg: utl.json_value_or_default(skater, Keys.faceoff_winning_pctg),
                Keys.toi: utl.json_value_or_default(skater, Keys.toi),
                Keys.blocked_shots: utl.json_value_or_default(skater, Keys.blocked_shots),
                Keys.shifts: utl.json_value_or_default(skater, Keys.shifts),
                Keys.giveaways: utl.json_value_or_default(skater, Keys.giveaways),
                Keys.takeaways: utl.json_value_or_default(skater, Keys.takeaways),
                Keys.team_id: team_id,
                Keys.team_role: team_role.value
            }

        meta_db[DB.skater_stats_table_name] = {
            Keys.last_update: datetime.now(timezone.utc)
        }
        logger.info("Finished adding skaters to database.")

    @staticmethod
    def _process_goalies(
        goalies: Dict[str, object],
        data: Dict[str, SqliteDict],
        game_id: str,
        team_id: str,
        team_role: HomeOrAway
    ) -> None:
        """Process the goalies in the provided data.

        Args:
            goalies (Dict[str, object]): Goalie JSON data.
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data in.
            game_id (str): Game ID for the game represented in the data.
            team_id (str): Team ID for the team represented in the data.
            team_role (HomeOrAway): Team role represented in the data.
        """
        logger.info("Started adding goalies to database.")
        goalie_stats_db = data[DB.goalie_stats_table_name]
        meta_db = data[DB.meta_table_name]

        for goalie in goalies:
            logger.info(f"Processing goalie. Goalie:'{goalie}'.")
            goalie_stats_db[len(goalie_stats_db)+1] = {
                Keys.game_id: game_id,
                Keys.player_id: utl.json_value_or_default(goalie, Keys.player_id),
                Keys.even_strength_shots_against: utl.json_value_or_default(goalie, Keys.even_strength_shots_against),
                Keys.power_play_shots_against: utl.json_value_or_default(goalie, Keys.power_play_shots_against),
                Keys.shorthanded_shots_against: utl.json_value_or_default(goalie, Keys.power_play_shots_against),
                Keys.save_shots_against: utl.json_value_or_default(goalie, Keys.save_shots_against),
                Keys.save_pctg: utl.json_value_or_default(goalie, Keys.save_pctg),
                Keys.even_strength_goals_against: utl.json_value_or_default(goalie, Keys.even_strength_goals_against),
                Keys.power_play_goals_against: utl.json_value_or_default(goalie, Keys.power_play_goals_against),
                Keys.shorthanded_goals_against: utl.json_value_or_default(goalie, Keys.shorthanded_goals_against),
                Keys.pim: utl.json_value_or_default(goalie, Keys.pim),
                Keys.goals_against: utl.json_value_or_default(goalie, Keys.goals_against),
                Keys.toi: utl.json_value_or_default(goalie, Keys.toi),
                Keys.starter: utl.json_value_or_default(goalie, Keys.starter),
                Keys.decision: utl.json_value_or_default(goalie, Keys.decision),
                Keys.shots_against: utl.json_value_or_default(goalie, Keys.shots_against),
                Keys.saves: utl.json_value_or_default(goalie, Keys.saves),
                Keys.team_id: team_id,
                Keys.team_role: team_role.value
            }

        meta_db[DB.goalie_stats_table_name] = {
            Keys.last_update: datetime.now(timezone.utc)
        }
        logger.info("Finished adding goalies to database.")

    @staticmethod
    def populate_players(
        data: Dict[str, SqliteDict]
    ) -> None:
        """Populate the players into the players table.

        Args:
            data (Dict[str, SqliteDict]): Dictionary of tables to store raw data in.
        """
        logger.info("Started adding players to database.")
        players_db = data[DB.players_table_name]
        meta_db = data[DB.meta_table_name]
        
        # I don't find this endpoint in the nhlpy APIs. Making a manual request
        # to get all active players.
        url = "https://search.d3.nhle.com/api/v1/search/player"
        params = dict(
            culture="en-us",
            limit="50000",
            q="*",
            active="true"
        )
        resp = requests.get(url=url, params=params)
        players_json = resp.json()

        if not players_json:
            logger.error("Unable to get JSON response content from players query.")
        else:
            for player in players_json:
                logger.info(f"Processing player: Player: '{player}'.")
                player_id = utl.json_value_or_default(player, Keys.player_id, default=None)
                stats = execution_context.client.stats.player_career_stats(player_id)
                first_name = utl.json_value_or_default(stats, Keys.first_name, Keys.default, default="")
                last_name = utl.json_value_or_default(stats, Keys.last_name, Keys.default, default="")
                players_db[player_id] = {
                    Keys.current_team_id: utl.json_value_or_default(stats, Keys.current_team_id),
                    Keys.first_name: first_name,
                    Keys.last_name: last_name,
                    Keys.height_in_cm: utl.json_value_or_default(stats, Keys.height_in_cm),
                    Keys.weight_in_kg: utl.json_value_or_default(stats, Keys.weight_in_kg)
                }
                logger.info(f"Added player '{last_name}, {first_name}' to players table.")

            meta_db[DB.players_table_name] = {
                Keys.last_update: datetime.now(timezone.utc)
            }
        logger.info("Finished adding players to database.")


class Database:
    """Constants related to databases
    """
    # Database table names
    skater_stats_table_name = "skater_stats"
    goalie_stats_table_name = "goalie_stats"
    players_table_name = "players"
    games_table_name = "games"
    meta_table_name = "meta"


from typing import List

from daterangeparser import parse as drp
from datetime import datetime, timedelta
import dateutil.parser as parser

from nhl_predictor.model.algorithms import Algorithms
from nhl_predictor.predictor.linear_regression import PredictLinearRegression
from nhl_predictor.shared.constants.json import JSON as Keys
from nhl_predictor.shared.execution_context import ExecutionContext
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.shared.utility import Utility as utl

logger = LoggingConfig.get_logger(__name__)
execution_context = ExecutionContext()

class Predictor:
    """Class to manage prediction operations.
    """
    
    def predict_by_date(
        algorithm: Algorithms,
        date: str = None,
        date_range: str = None,
    ) -> None:
        """Start a prediction based on a date or dates.

        Args:
            algorithm (Algorithms): Which ML algorithm to use.
            date (str, optional): Single date filter. Defaults to None.
            date_range (str, optional): Date range filter. Defaults to None.
        """
        games = Predictor._get_games(date, date_range)
        Predictor._predict(algorithm, games)
    
    @staticmethod
    def predict_single_game(
        algorithm: Algorithms,
        game_id: str
    ) -> None:
        """Starts a prediction based on a specific game ID.

        Args:
            algorithm (Algorithms): Which ML algorithm to use.
            game_id (str): The ID of the game to predict.
        """
        game = Predictor._get_game_by_id(game_id)
        Predictor._predict(algorithm, [game])
    
    @staticmethod
    def _predict(
        algorithm: Algorithms,
        games: List[int]
    ) -> None:
        """Core predict method

        Args:
            algorithm (Algorithms): Which ML algorithm to use.
            games (List[int]): List of game IDs to predict for.
        """
        match algorithm:
            case Algorithms.linear_regression:
                PredictLinearRegression.predict(games)
            case _:
                logger.error("Invalid algorithm provided to predict.")
    
    @staticmethod
    def list_games(
        date: str,
        date_range: str
    ) -> None:
        """Print out a report of games on a given date or during a given date range.

        Args:
            date (str): Single date filter.
            date_range (str): Date range filter.
        """
        table = []
        games = Predictor._get_games(date, date_range)
        table.append([Keys.game_id, Keys.away_team, Keys.home_team, Keys.game_state])
        for game in games:
            table.append([
                str(utl.json_value_or_default(game, Keys.id)),
                str(utl.json_value_or_default(game, Keys.away_team, Keys.common_name, Keys.default)),
                str(utl.json_value_or_default(game, Keys.home_team, Keys.common_name, Keys.default)),
                str(utl.json_value_or_default(game, Keys.game_state))
            ])
        utl.print_table(table, hasHeader=True)

    @staticmethod
    def _parse_date_range(
        date_range: str
    ) -> tuple[datetime, datetime]:
        """Parse a date range string into starting and ending datetime instances.

        Args:
            date_range (str): Date range to be parsed.

        Returns:
            tuple[datetime, datetime] Tuple with start and end datetime instances.
        """
        date_range_start = date_range_end = None
        if date_range is not None:
            date_range_start, date_range_end = drp(date_range)
        return date_range_start, date_range_end
    
    @staticmethod
    def _get_games(
        date: str,
        date_range: str
    ) -> List[object]:
        """Get set of games schduled on a given date or during a given date range.

        Args:
            date (str): Single date filter.
            date_range (str): Date range filter.

        Returns:
            List[object]: List of games matching provided date filters.
        """
        date_range_start, date_range_end = Predictor._parse_date_range(date_range)
        if date is not None:
            date = parser.parse(date)
            return Predictor._get_games_for_date(date)
        elif date_range_start is not None and date_range_end is not None:
            return Predictor._get_games_for_date_range(date_range_start, date_range_end)
        else:
            # TODO: Throw?
            logger.error("No valid date option supplied")
            return

    @staticmethod
    def _get_games_for_date_range(
        date_range_start: datetime,
        date_range_end: datetime
    ) -> List[object]:
        """Get the set of games scheduled during a given date range.

        Args:
            date_range_start (datetime): Start of the date range filter.
            date_range_end (datetime): End of the date range filter.

        Returns:
            List[object]: List of game JSON.
        """
        games = []
        number_of_days = (date_range_end - date_range_start).days + 1
        date_list = [date_range_end - timedelta(days=x) for x in range(number_of_days)]
        for date in date_list:
            next_games = Predictor._get_games_for_date(date)
            games.extend(next_games)
        return games

    @staticmethod
    def _get_games_for_date(date: datetime) -> List[object]:
        """Get the set of games scheduled on the given date.

        Args:
            date (datetime): Date filter for games.

        Returns:
            List[object]: List of game JSON.
        """
        schedule = execution_context.client.schedule.daily_schedule(str(date)[:10])
        return schedule["games"]
    
    @staticmethod
    def _get_game_by_id(id: int) -> List[object]:
        """Get game by game ID

        Args:
            id (int): The game ID for the desired game.

        Returns:
            List[object]: List of game JSON.
        """
        return execution_context.client.game_center.boxscore(id)


from enum import Enum


class GameState(str, Enum):
    """Enum mapping game type labels to the numerical value used by the
    NHL API to indicate game type.

    Since the NHL doesn't document their API, value map was obtained
    from here:
    https://github.com/Zmalski/NHL-API-Reference/issues/23#issuecomment-2492925102
    """
    Future = "FUT"
    Pregame = "PRE"
    Live = "LIVE"
    Final = "FINAL"
    Official = "OFF"

# These are the GameStates that indicate player stats are recorded.
GameStatesForDataset = [
    GameState.Official
]

# These are the GameStates where the roster has already been added.
GameStatesForPrediction = [
    GameState.Live,
    GameState.Final,
    GameState.Official
]


import os
from pickle import dump

import numpy as np
import statsmodels.api as sm
from ansimarkup import ansiprint as print
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from nhl_predictor.model.summarizer_manager import SummarizerTypes
from nhl_predictor.shared.constants.database import Database as DB
from nhl_predictor.shared.execution_context import ExecutionContext
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.shared.utility import Utility as utl

logger = LoggingConfig.get_logger(__name__)
execution_context = ExecutionContext()
_model_filename_part = "LinearRegression"

class TrainLinearRegression:
    """Static class to train a linear regression model on NHL data.
    """
    
    @staticmethod
    def train() -> None:
        """Entry point for linear regression model training.
        """
        logger.info("Start of model training.")
        
        data = utl.get_pandas_tables(
            DB.players_table_name,
            DB.skater_stats_table_name,
            DB.goalie_stats_table_name,
            DB.games_table_name,
            DB.meta_table_name,
            path=execution_context.app_dir
        )

        summarizer = SummarizerTypes.get_summarizer(execution_context.summarizer_type)
        game_data = summarizer.summarize(data)

        train, test = train_test_split(game_data, test_size=0.2)

        # TODO: This assumes that the last column is our winner column.  The
        # summarizer needs to either guarantee this or return them separately.
        targetsColumnName = train.columns[len(train.columns)-1]
        train_targets = train[targetsColumnName].to_list()
        train.drop(targetsColumnName, axis=1, inplace=True)
        targetsColumnName = test.columns[len(test.columns)-1]
        test_targets = test[targetsColumnName].to_list()
        test.drop(targetsColumnName, axis=1, inplace=True)

        # create LR model
        model = LinearRegression()

        # Fit
        model.fit(train, train_targets)

        summary_table = []
        pvalue_table = []

        # r-squared
        train_pred = model.predict(train)
        r_squared = r2_score(train_targets, train_pred)
        summary_table.append(["R-squared", str(r_squared)])

        # P-Values
        x_intercept = sm.add_constant(train)
        model_sm = sm.OLS(train_targets, x_intercept).fit()
        pvalue_table = list(map(list, model_sm.pvalues.items()))
        for idx in range(len(pvalue_table)):
            pvalue_table[idx][1] = np.format_float_scientific(pvalue_table[idx][1])
        
        summary_table.append([
            "Mean squared error:",
            "%.2f" % mean_squared_error(train_targets, train_pred)
        ])

        print("<green>Stats from train operation:</green>")
        utl.print_table(summary_table)
        print("<blue>PValues:</blue>")
        utl.print_table(pvalue_table)
        
        summary_table = []
        pvalue_table = []

        test_pred = model.predict(test)
        r_squared = r2_score(test_targets, test_pred)
        summary_table.append(["R-squared: ", str(r_squared)])

        # P-Values
        x_intercept = sm.add_constant(test)
        model_sm = sm.OLS(test_targets, x_intercept).fit()
        pvalue_table = list(map(list, model_sm.pvalues.items()))
        for idx in range(len(pvalue_table)):
            pvalue_table[idx][1] = np.format_float_scientific(pvalue_table[idx][1])

        summary_table.append([
            "Mean squared error:",
            "%.2f" % mean_squared_error(test_targets, test_pred)
        ])

        print("<green>Stats from test operation:</green>")
        utl.print_table(summary_table)
        print("<blue>PValues:</blue>")
        utl.print_table(pvalue_table)

        # Coefficients
        # metrics_table.append(["Coef: ", str(model.coef_)])
        # metrics_table.append(["Intercept: ", str(model.intercept_)])
        # print("Coef: ", model.coef_)
        # print("Intercept: ", model.intercept_)

        if execution_context.output_file:
            filename = execution_context.output_file
        else:
            filename = f"{summarizer.get_filename_prefix()}_{_model_filename_part}.pkl"

        with open(os.path.join(execution_context.app_dir, filename), "wb") as file:
            dump(model, file, protocol=5)

        logger.info("End of model training.")


from nhl_predictor.model.algorithms import Algorithms
from nhl_predictor.shared.logging_config import LoggingConfig
from nhl_predictor.trainer.linear_regression import TrainLinearRegression

logger = LoggingConfig.get_logger(__name__)

class Trainer:
    """Static class providing an entry point for training machine learning
    models.
    """

    @staticmethod
    def train(algorithm: Algorithms) -> None:
        """Train a model using the specified algorithm.

        Args:
            algorithm (Algorithms): The machine learning Algorithm to train a
            model for.
        """
        match algorithm:
            case Algorithms.linear_regression:
                TrainLinearRegression.train()
            case _:
                logger.error("Invalid algorithm provided to train.")


# Some APIs require the team abbreviation, others require
# the franchise ID. The main API that we're using takes the
# abbreviation, so we use that as our key and iterate. When
# needed, we can use the map to find the franchise ID to use
# when calling subsequent APIs.
TeamMap = {
    "MTL": 1,               # Canadiens
    "TOR": 5,               # Maple Leafs
    "BOS": 6,               # Bruins
    "NYR": 10,              # Rangers
    "CHI": 11,              # Blackhawks
    "DET": 12,              # Red Wings
    "LAK": 14,              # Kings
    "DAL": 15,              # Stars
    "PHI": 16,              # Flyers
    "PIT": 17,              # Penguins
    "STL": 18,              # Blues
    "BUF": 19,              # Sabres
    "VAN": 20,              # Canucks
    "CGY": 21,              # Flames
    "NYI": 22,              # Islanders
    "NJD": 23,              # Jersey Devils
    "WSH": 24,              # Capitals
    "EDM": 25,              # Oilers
    "CAR": 26,              # Hurricanes
    "COL": 27,              # Avalanche
    "SJS": 29,              # Sharks
    "OTT": 30,              # Senators
    "TBL": 31,              # Lightning
    "ANA": 32,              # Ducks
    "FLA": 33,              # Panthers
    "NSH": 34,              # Predators
    "WPG": 35,              # Jets
    "CBJ": 36,              # Blue Jackets
    "MIN": 37,              # Wild
    "VGK": 38,              # Golden Knights
    "SEA": 55,              # Kraken
    "UTA": 40,              # Mammoth
    "ARI": 40,              # Arizona -- Back compat for Mammoth
}